@@ -11,6 +11,7 @@ from pydantic import BaseModel, Field
 from mcp.server.fastmcp.exceptions import ToolError
 from mcp.server.fastmcp.utilities.context_injection import find_context_parameter
 from mcp.server.fastmcp.utilities.func_metadata import FuncMetadata, func_metadata
+from mcp.shared.tool_name_validation import validate_and_warn_tool_name
 from mcp.types import Icon, ToolAnnotations
 
 if TYPE_CHECKING:
@@ -34,6 +35,7 @@ class Tool(BaseModel):
     context_kwarg: str | None = Field(None, description="Name of the kwarg that should receive context")
     annotations: ToolAnnotations | None = Field(None, description="Optional annotations for the tool")
     icons: list[Icon] | None = Field(default=None, description="Optional list of icons for this tool")
+    meta: dict[str, Any] | None = Field(default=None, description="Optional metadata for this tool")
 
     @cached_property
     def output_schema(self) -> dict[str, Any] | None:
@@ -49,18 +51,21 @@ class Tool(BaseModel):
         context_kwarg: str | None = None,
         annotations: ToolAnnotations | None = None,
         icons: list[Icon] | None = None,
+        meta: dict[str, Any] | None = None,
         structured_output: bool | None = None,
     ) -> Tool:
         """Create a Tool from a function."""
         func_name = name or fn.__name__
 
+        validate_and_warn_tool_name(func_name)
+
         if func_name == "<lambda>":
             raise ValueError("You must provide a name for lambda functions")
 
         func_doc = description or fn.__doc__ or ""
         is_async = _is_async_callable(fn)
 
-        if context_kwarg is None:
+        if context_kwarg is None:  # pragma: no branch
             context_kwarg = find_context_parameter(fn)
 
         func_arg_metadata = func_metadata(
@@ -81,6 +86,7 @@ class Tool(BaseModel):
             context_kwarg=context_kwarg,
             annotations=annotations,
             icons=icons,
+            meta=meta,
         )
 
     async def run(
@@ -107,7 +113,7 @@ class Tool(BaseModel):
 
 
 def _is_async_callable(obj: Any) -> bool:
-    while isinstance(obj, functools.partial):
+    while isinstance(obj, functools.partial):  # pragma: no cover
         obj = obj.func
 
     return inspect.iscoroutinefunction(obj) or (
