@@ -15,6 +15,7 @@ from collections.abc import AsyncGenerator, Awaitable, Callable
 from contextlib import asynccontextmanager
 from dataclasses import dataclass
 from http import HTTPStatus
+from typing import Any
 
 import anyio
 from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
@@ -87,18 +88,18 @@ class EventStore(ABC):
     """
 
     @abstractmethod
-    async def store_event(self, stream_id: StreamId, message: JSONRPCMessage) -> EventId:
+    async def store_event(self, stream_id: StreamId, message: JSONRPCMessage | None) -> EventId:
         """
         Stores an event for later retrieval.
 
         Args:
             stream_id: ID of the stream the event belongs to
-            message: The JSON-RPC message to store
+            message: The JSON-RPC message to store, or None for priming events
 
         Returns:
             The generated event ID for the stored event
         """
-        pass
+        pass  # pragma: no cover
 
     @abstractmethod
     async def replay_events_after(
@@ -116,7 +117,7 @@ class EventStore(ABC):
         Returns:
             The stream ID of the replayed events
         """
-        pass
+        pass  # pragma: no cover
 
 
 class StreamableHTTPServerTransport:
@@ -140,6 +141,7 @@ class StreamableHTTPServerTransport:
         is_json_response_enabled: bool = False,
         event_store: EventStore | None = None,
         security_settings: TransportSecuritySettings | None = None,
+        retry_interval: int | None = None,
     ) -> None:
         """
         Initialize a new StreamableHTTP server transport.
@@ -153,6 +155,10 @@ class StreamableHTTPServerTransport:
                         resumability will be enabled, allowing clients to
                         reconnect and resume messages.
             security_settings: Optional security settings for DNS rebinding protection.
+            retry_interval: Retry interval in milliseconds to suggest to clients in SSE
+                           retry field. When set, the server will send a retry field in
+                           SSE priming events to control client reconnection timing for
+                           polling behavior. Only used when event_store is provided.
 
         Raises:
             ValueError: If the session ID contains invalid characters.
@@ -164,6 +170,7 @@ class StreamableHTTPServerTransport:
         self.is_json_response_enabled = is_json_response_enabled
         self._event_store = event_store
         self._security = TransportSecurityMiddleware(security_settings)
+        self._retry_interval = retry_interval
         self._request_streams: dict[
             RequestId,
             tuple[
@@ -171,6 +178,7 @@ class StreamableHTTPServerTransport:
                 MemoryObjectReceiveStream[EventMessage],
             ],
         ] = {}
+        self._sse_stream_writers: dict[RequestId, MemoryObjectSendStream[dict[str, str]]] = {}
         self._terminated = False
 
     @property
@@ -178,6 +186,91 @@ class StreamableHTTPServerTransport:
         """Check if this transport has been explicitly terminated."""
         return self._terminated
 
+    def close_sse_stream(self, request_id: RequestId) -> None:  # pragma: no cover
+        """Close SSE connection for a specific request without terminating the stream.
+
+        This method closes the HTTP connection for the specified request, triggering
+        client reconnection. Events continue to be stored in the event store and will
+        be replayed when the client reconnects with Last-Event-ID.
+
+        Use this to implement polling behavior during long-running operations -
+        client will reconnect after the retry interval specified in the priming event.
+
+        Args:
+            request_id: The request ID whose SSE stream should be closed.
+
+        Note:
+            This is a no-op if there is no active stream for the request ID.
+            Requires event_store to be configured for events to be stored during
+            the disconnect.
+        """
+        writer = self._sse_stream_writers.pop(request_id, None)
+        if writer:
+            writer.close()
+
+        # Also close and remove request streams
+        if request_id in self._request_streams:
+            send_stream, receive_stream = self._request_streams.pop(request_id)
+            send_stream.close()
+            receive_stream.close()
+
+    def close_standalone_sse_stream(self) -> None:  # pragma: no cover
+        """Close the standalone GET SSE stream, triggering client reconnection.
+
+        This method closes the HTTP connection for the standalone GET stream used
+        for unsolicited server-to-client notifications. The client SHOULD reconnect
+        with Last-Event-ID to resume receiving notifications.
+
+        Use this to implement polling behavior for the notification stream -
+        client will reconnect after the retry interval specified in the priming event.
+
+        Note:
+            This is a no-op if there is no active standalone SSE stream.
+            Requires event_store to be configured for events to be stored during
+            the disconnect.
+            Currently, client reconnection for standalone GET streams is NOT
+            implemented - this is a known gap (see test_standalone_get_stream_reconnection).
+        """
+        self.close_sse_stream(GET_STREAM_KEY)
+
+    def _create_session_message(  # pragma: no cover
+        self,
+        message: JSONRPCMessage,
+        request: Request,
+        request_id: RequestId,
+    ) -> SessionMessage:
+        """Create a session message with metadata including close_sse_stream callback."""
+
+        async def close_stream_callback() -> None:
+            self.close_sse_stream(request_id)
+
+        async def close_standalone_stream_callback() -> None:
+            self.close_standalone_sse_stream()
+
+        metadata = ServerMessageMetadata(
+            request_context=request,
+            close_sse_stream=close_stream_callback,
+            close_standalone_sse_stream=close_standalone_stream_callback,
+        )
+        return SessionMessage(message, metadata=metadata)
+
+    async def _send_priming_event(  # pragma: no cover
+        self,
+        request_id: RequestId,
+        sse_stream_writer: MemoryObjectSendStream[dict[str, Any]],
+    ) -> None:
+        """Send priming event for SSE resumability if event_store is configured."""
+        if not self._event_store:
+            return
+        priming_event_id = await self._event_store.store_event(
+            str(request_id),  # Convert RequestId to StreamId (str)
+            None,  # Priming event has no payload
+        )
+        priming_event: dict[str, str | int] = {"id": priming_event_id, "data": ""}
+        if self._retry_interval is not None:
+            priming_event["retry"] = self._retry_interval
+        await sse_stream_writer.send(priming_event)
+
     def _create_error_response(
         self,
         error_message: str,
@@ -187,7 +280,7 @@ class StreamableHTTPServerTransport:
     ) -> Response:
         """Create an error response with a simple string message."""
         response_headers = {"Content-Type": CONTENT_TYPE_JSON}
-        if headers:
+        if headers:  # pragma: no cover
             response_headers.update(headers)
 
         if self.mcp_session_id:
@@ -209,7 +302,7 @@ class StreamableHTTPServerTransport:
             headers=response_headers,
         )
 
-    def _create_json_response(
+    def _create_json_response(  # pragma: no cover
         self,
         response_message: JSONRPCMessage | None,
         status_code: HTTPStatus = HTTPStatus.OK,
@@ -229,11 +322,11 @@ class StreamableHTTPServerTransport:
             headers=response_headers,
         )
 
-    def _get_session_id(self, request: Request) -> str | None:
+    def _get_session_id(self, request: Request) -> str | None:  # pragma: no cover
         """Extract the session ID from request headers."""
         return request.headers.get(MCP_SESSION_ID_HEADER)
 
-    def _create_event_data(self, event_message: EventMessage) -> dict[str, str]:
+    def _create_event_data(self, event_message: EventMessage) -> dict[str, str]:  # pragma: no cover
         """Create event data dictionary from an EventMessage."""
         event_data = {
             "event": "message",
@@ -246,7 +339,7 @@ class StreamableHTTPServerTransport:
 
         return event_data
 
-    async def _clean_up_memory_streams(self, request_id: RequestId) -> None:
+    async def _clean_up_memory_streams(self, request_id: RequestId) -> None:  # pragma: no cover
         """Clean up memory streams for a given request ID."""
         if request_id in self._request_streams:
             try:
@@ -267,11 +360,11 @@ class StreamableHTTPServerTransport:
         # Validate request headers for DNS rebinding protection
         is_post = request.method == "POST"
         error_response = await self._security.validate_request(request, is_post=is_post)
-        if error_response:
+        if error_response:  # pragma: no cover
             await error_response(scope, receive, send)
             return
 
-        if self._terminated:
+        if self._terminated:  # pragma: no cover
             # If the session has been terminated, return 404 Not Found
             response = self._create_error_response(
                 "Not Found: Session has been terminated",
@@ -282,11 +375,11 @@ class StreamableHTTPServerTransport:
 
         if request.method == "POST":
             await self._handle_post_request(scope, request, receive, send)
-        elif request.method == "GET":
+        elif request.method == "GET":  # pragma: no cover
             await self._handle_get_request(request, send)
-        elif request.method == "DELETE":
+        elif request.method == "DELETE":  # pragma: no cover
             await self._handle_delete_request(request, send)
-        else:
+        else:  # pragma: no cover
             await self._handle_unsupported_request(request, send)
 
     def _check_accept_headers(self, request: Request) -> tuple[bool, bool]:
@@ -306,24 +399,40 @@ class StreamableHTTPServerTransport:
 
         return any(part == CONTENT_TYPE_JSON for part in content_type_parts)
 
+    async def _validate_accept_header(self, request: Request, scope: Scope, send: Send) -> bool:  # pragma: no cover
+        """Validate Accept header based on response mode. Returns True if valid."""
+        has_json, has_sse = self._check_accept_headers(request)
+        if self.is_json_response_enabled:
+            # For JSON-only responses, only require application/json
+            if not has_json:
+                response = self._create_error_response(
+                    "Not Acceptable: Client must accept application/json",
+                    HTTPStatus.NOT_ACCEPTABLE,
+                )
+                await response(scope, request.receive, send)
+                return False
+        # For SSE responses, require both content types
+        elif not (has_json and has_sse):
+            response = self._create_error_response(
+                "Not Acceptable: Client must accept both application/json and text/event-stream",
+                HTTPStatus.NOT_ACCEPTABLE,
+            )
+            await response(scope, request.receive, send)
+            return False
+        return True
+
     async def _handle_post_request(self, scope: Scope, request: Request, receive: Receive, send: Send) -> None:
         """Handle POST requests containing JSON-RPC messages."""
         writer = self._read_stream_writer
-        if writer is None:
+        if writer is None:  # pragma: no cover
             raise ValueError("No read stream writer available. Ensure connect() is called first.")
         try:
-            # Check Accept headers
-            has_json, has_sse = self._check_accept_headers(request)
-            if not (has_json and has_sse):
-                response = self._create_error_response(
-                    ("Not Acceptable: Client must accept both application/json and text/event-stream"),
-                    HTTPStatus.NOT_ACCEPTABLE,
-                )
-                await response(scope, receive, send)
+            # Validate Accept header
+            if not await self._validate_accept_header(request, scope, send):
                 return
 
             # Validate Content-Type
-            if not self._check_content_type(request):
+            if not self._check_content_type(request):  # pragma: no cover
                 response = self._create_error_response(
                     "Unsupported Media Type: Content-Type must be application/json",
                     HTTPStatus.UNSUPPORTED_MEDIA_TYPE,
@@ -341,9 +450,9 @@ class StreamableHTTPServerTransport:
                 await response(scope, receive, send)
                 return
 
-            try:
+            try:  # pragma: no cover
                 message = JSONRPCMessage.model_validate(raw_message)
-            except ValidationError as e:
+            except ValidationError as e:  # pragma: no cover
                 response = self._create_error_response(
                     f"Validation error: {str(e)}",
                     HTTPStatus.BAD_REQUEST,
@@ -353,9 +462,11 @@ class StreamableHTTPServerTransport:
                 return
 
             # Check if this is an initialization request
-            is_initialization_request = isinstance(message.root, JSONRPCRequest) and message.root.method == "initialize"
+            is_initialization_request = (
+                isinstance(message.root, JSONRPCRequest) and message.root.method == "initialize"
+            )  # pragma: no cover
 
-            if is_initialization_request:
+            if is_initialization_request:  # pragma: no cover
                 # Check if the server already has an established session
                 if self.mcp_session_id:
                     # Check if request has a session ID
@@ -369,11 +480,11 @@ class StreamableHTTPServerTransport:
                         )
                         await response(scope, receive, send)
                         return
-            elif not await self._validate_request_headers(request, send):
+            elif not await self._validate_request_headers(request, send):  # pragma: no cover
                 return
 
             # For notifications and responses only, return 202 Accepted
-            if not isinstance(message.root, JSONRPCRequest):
+            if not isinstance(message.root, JSONRPCRequest):  # pragma: no cover
                 # Create response object and send it
                 response = self._create_json_response(
                     None,
@@ -389,12 +500,12 @@ class StreamableHTTPServerTransport:
                 return
 
             # Extract the request ID outside the try block for proper scope
-            request_id = str(message.root.id)
+            request_id = str(message.root.id)  # pragma: no cover
             # Register this stream for the request ID
-            self._request_streams[request_id] = anyio.create_memory_object_stream[EventMessage](0)
-            request_stream_reader = self._request_streams[request_id][1]
+            self._request_streams[request_id] = anyio.create_memory_object_stream[EventMessage](0)  # pragma: no cover
+            request_stream_reader = self._request_streams[request_id][1]  # pragma: no cover
 
-            if self.is_json_response_enabled:
+            if self.is_json_response_enabled:  # pragma: no cover
                 # Process the message
                 metadata = ServerMessageMetadata(request_context=request)
                 session_message = SessionMessage(message, metadata=metadata)
@@ -437,14 +548,20 @@ class StreamableHTTPServerTransport:
                     await response(scope, receive, send)
                 finally:
                     await self._clean_up_memory_streams(request_id)
-            else:
+            else:  # pragma: no cover
                 # Create SSE stream
                 sse_stream_writer, sse_stream_reader = anyio.create_memory_object_stream[dict[str, str]](0)
 
+                # Store writer reference so close_sse_stream() can close it
+                self._sse_stream_writers[request_id] = sse_stream_writer
+
                 async def sse_writer():
                     # Get the request ID from the incoming request message
                     try:
                         async with sse_stream_writer, request_stream_reader:
+                            # Send priming event for SSE resumability
+                            await self._send_priming_event(request_id, sse_stream_writer)
+
                             # Process messages from the request-specific stream
                             async for event_message in request_stream_reader:
                                 # Build the event data
@@ -457,10 +574,14 @@ class StreamableHTTPServerTransport:
                                     JSONRPCResponse | JSONRPCError,
                                 ):
                                     break
+                    except anyio.ClosedResourceError:
+                        # Expected when close_sse_stream() is called
+                        logger.debug("SSE stream closed by close_sse_stream()")
                     except Exception:
                         logger.exception("Error in SSE writer")
                     finally:
                         logger.debug("Closing SSE writer")
+                        self._sse_stream_writers.pop(request_id, None)
                         await self._clean_up_memory_streams(request_id)
 
                 # Create and start EventSourceResponse
@@ -484,8 +605,7 @@ class StreamableHTTPServerTransport:
                     async with anyio.create_task_group() as tg:
                         tg.start_soon(response, scope, receive, send)
                         # Then send the message to be processed by the server
-                        metadata = ServerMessageMetadata(request_context=request)
-                        session_message = SessionMessage(message, metadata=metadata)
+                        session_message = self._create_session_message(message, request, request_id)
                         await writer.send(session_message)
                 except Exception:
                     logger.exception("SSE response error")
@@ -493,7 +613,7 @@ class StreamableHTTPServerTransport:
                     await sse_stream_reader.aclose()
                     await self._clean_up_memory_streams(request_id)
 
-        except Exception as err:
+        except Exception as err:  # pragma: no cover
             logger.exception("Error handling POST request")
             response = self._create_error_response(
                 f"Error handling POST request: {err}",
@@ -505,7 +625,7 @@ class StreamableHTTPServerTransport:
                 await writer.send(Exception(err))
             return
 
-    async def _handle_get_request(self, request: Request, send: Send) -> None:
+    async def _handle_get_request(self, request: Request, send: Send) -> None:  # pragma: no cover
         """
         Handle GET request to establish SSE.
 
@@ -597,7 +717,7 @@ class StreamableHTTPServerTransport:
             await sse_stream_reader.aclose()
             await self._clean_up_memory_streams(GET_STREAM_KEY)
 
-    async def _handle_delete_request(self, request: Request, send: Send) -> None:
+    async def _handle_delete_request(self, request: Request, send: Send) -> None:  # pragma: no cover
         """Handle DELETE requests for explicit session termination."""
         # Validate session ID
         if not self.mcp_session_id:
@@ -633,25 +753,25 @@ class StreamableHTTPServerTransport:
         request_stream_keys = list(self._request_streams.keys())
 
         # Close all request streams asynchronously
-        for key in request_stream_keys:
+        for key in request_stream_keys:  # pragma: no cover
             await self._clean_up_memory_streams(key)
 
         # Clear the request streams dictionary immediately
         self._request_streams.clear()
         try:
-            if self._read_stream_writer is not None:
+            if self._read_stream_writer is not None:  # pragma: no branch
                 await self._read_stream_writer.aclose()
-            if self._read_stream is not None:
+            if self._read_stream is not None:  # pragma: no branch
                 await self._read_stream.aclose()
-            if self._write_stream_reader is not None:
+            if self._write_stream_reader is not None:  # pragma: no branch
                 await self._write_stream_reader.aclose()
-            if self._write_stream is not None:
+            if self._write_stream is not None:  # pragma: no branch
                 await self._write_stream.aclose()
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             # During cleanup, we catch all exceptions since streams might be in various states
             logger.debug(f"Error closing streams: {e}")
 
-    async def _handle_unsupported_request(self, request: Request, send: Send) -> None:
+    async def _handle_unsupported_request(self, request: Request, send: Send) -> None:  # pragma: no cover
         """Handle unsupported HTTP methods."""
         headers = {
             "Content-Type": CONTENT_TYPE_JSON,
@@ -667,14 +787,14 @@ class StreamableHTTPServerTransport:
         )
         await response(request.scope, request.receive, send)
 
-    async def _validate_request_headers(self, request: Request, send: Send) -> bool:
+    async def _validate_request_headers(self, request: Request, send: Send) -> bool:  # pragma: no cover
         if not await self._validate_session(request, send):
             return False
         if not await self._validate_protocol_version(request, send):
             return False
         return True
 
-    async def _validate_session(self, request: Request, send: Send) -> bool:
+    async def _validate_session(self, request: Request, send: Send) -> bool:  # pragma: no cover
         """Validate the session ID in the request."""
         if not self.mcp_session_id:
             # If we're not using session IDs, return True
@@ -703,7 +823,7 @@ class StreamableHTTPServerTransport:
 
         return True
 
-    async def _validate_protocol_version(self, request: Request, send: Send) -> bool:
+    async def _validate_protocol_version(self, request: Request, send: Send) -> bool:  # pragma: no cover
         """Validate the protocol version header in the request."""
         # Get the protocol version from the request headers
         protocol_version = request.headers.get(MCP_PROTOCOL_VERSION_HEADER)
@@ -725,7 +845,7 @@ class StreamableHTTPServerTransport:
 
         return True
 
-    async def _replay_events(self, last_event_id: str, request: Request, send: Send) -> None:
+    async def _replay_events(self, last_event_id: str, request: Request, send: Send) -> None:  # pragma: no cover
         """
         Replays events that would have been sent after the specified event ID.
         Only used when resumability is enabled.
@@ -760,6 +880,13 @@ class StreamableHTTPServerTransport:
 
                         # If stream ID not in mapping, create it
                         if stream_id and stream_id not in self._request_streams:
+                            # Register SSE writer so close_sse_stream() can close it
+                            self._sse_stream_writers[stream_id] = sse_stream_writer
+
+                            # Send priming event for this new connection
+                            await self._send_priming_event(stream_id, sse_stream_writer)
+
+                            # Create new request streams for this connection
                             self._request_streams[stream_id] = anyio.create_memory_object_stream[EventMessage](0)
                             msg_reader = self._request_streams[stream_id][1]
 
@@ -769,6 +896,9 @@ class StreamableHTTPServerTransport:
                                     event_data = self._create_event_data(event_message)
 
                                     await sse_stream_writer.send(event_data)
+                except anyio.ClosedResourceError:
+                    # Expected when close_sse_stream() is called
+                    logger.debug("Replay SSE stream closed by close_sse_stream()")
                 except Exception:
                     logger.exception("Error in replay sender")
 
@@ -826,7 +956,7 @@ class StreamableHTTPServerTransport:
         # Start a task group for message routing
         async with anyio.create_task_group() as tg:
             # Create a message router that distributes messages to request streams
-            async def message_router():
+            async def message_router():  # pragma: no cover
                 try:
                     async for session_message in write_stream_reader:
                         # Determine which request stream(s) should receive this message
@@ -885,7 +1015,7 @@ class StreamableHTTPServerTransport:
                 # Yield the streams for the caller to use
                 yield read_stream, write_stream
             finally:
-                for stream_id in list(self._request_streams.keys()):
+                for stream_id in list(self._request_streams.keys()):  # pragma: no cover
                     await self._clean_up_memory_streams(stream_id)
                 self._request_streams.clear()
 
@@ -895,6 +1025,6 @@ class StreamableHTTPServerTransport:
                     await read_stream.aclose()
                     await write_stream_reader.aclose()
                     await write_stream.aclose()
-                except Exception as e:
+                except Exception as e:  # pragma: no cover
                     # During cleanup, we catch all exceptions since streams might be in various states
                     logger.debug(f"Error closing streams: {e}")
