@@ -14,6 +14,9 @@ ResumptionToken = str
 
 ResumptionTokenUpdateCallback = Callable[[ResumptionToken], Awaitable[None]]
 
+# Callback type for closing SSE streams without terminating
+CloseSSEStreamCallback = Callable[[], Awaitable[None]]
+
 
 @dataclass
 class ClientMessageMetadata:
@@ -30,6 +33,10 @@ class ServerMessageMetadata:
     related_request_id: RequestId | None = None
     # Request-specific context (e.g., headers, auth info)
     request_context: object | None = None
+    # Callback to close SSE stream for the current request without terminating
+    close_sse_stream: CloseSSEStreamCallback | None = None
+    # Callback to close the standalone GET SSE stream (for unsolicited notifications)
+    close_standalone_sse_stream: CloseSSEStreamCallback | None = None
 
 
 MessageMetadata = ClientMessageMetadata | ServerMessageMetadata | None
