@@ -5,7 +5,12 @@ import pytest
 
 import mcp
 from mcp import types
-from mcp.client.session_group import ClientSessionGroup, SseServerParameters, StreamableHttpParameters
+from mcp.client.session_group import (
+    ClientSessionGroup,
+    ClientSessionParameters,
+    SseServerParameters,
+    StreamableHttpParameters,
+)
 from mcp.client.stdio import StdioServerParameters
 from mcp.shared.exceptions import McpError
 
@@ -50,7 +55,7 @@ class TestClientSessionGroup:
         mock_session = mock.AsyncMock()
 
         # --- Prepare Session Group ---
-        def hook(name: str, server_info: types.Implementation) -> str:
+        def hook(name: str, server_info: types.Implementation) -> str:  # pragma: no cover
             return f"{(server_info.name)}-{name}"
 
         mcp_session_group = ClientSessionGroup(component_name_hook=hook)
@@ -62,7 +67,7 @@ class TestClientSessionGroup:
         # --- Test Execution ---
         result = await mcp_session_group.call_tool(
             name="server1-my_tool",
-            args={
+            arguments={
                 "name": "value1",
                 "args": {},
             },
@@ -73,6 +78,9 @@ class TestClientSessionGroup:
         mock_session.call_tool.assert_called_once_with(
             "my_tool",
             {"name": "value1", "args": {}},
+            read_timeout_seconds=None,
+            progress_callback=None,
+            meta=None,
         )
 
     async def test_connect_to_server(self, mock_exit_stack: contextlib.AsyncExitStack):
@@ -329,7 +337,7 @@ class TestClientSessionGroup:
                     (
                         returned_server_info,
                         returned_session,
-                    ) = await group._establish_session(server_params_instance)
+                    ) = await group._establish_session(server_params_instance, ClientSessionParameters())
 
                 # --- Assertions ---
                 # 1. Assert the correct specific client function was called
@@ -344,7 +352,7 @@ class TestClientSessionGroup:
                         timeout=server_params_instance.timeout,
                         sse_read_timeout=server_params_instance.sse_read_timeout,
                     )
-                elif client_type_name == "streamablehttp":
+                elif client_type_name == "streamablehttp":  # pragma: no branch
                     assert isinstance(server_params_instance, StreamableHttpParameters)
                     mock_specific_client_func.assert_called_once_with(
                         url=server_params_instance.url,
@@ -357,7 +365,17 @@ class TestClientSessionGroup:
                 mock_client_cm_instance.__aenter__.assert_awaited_once()
 
                 # 2. Assert ClientSession was called correctly
-                mock_ClientSession_class.assert_called_once_with(mock_read_stream, mock_write_stream)
+                mock_ClientSession_class.assert_called_once_with(
+                    mock_read_stream,
+                    mock_write_stream,
+                    read_timeout_seconds=None,
+                    sampling_callback=None,
+                    elicitation_callback=None,
+                    list_roots_callback=None,
+                    logging_callback=None,
+                    message_handler=None,
+                    client_info=None,
+                )
                 mock_raw_session_cm.__aenter__.assert_awaited_once()
                 mock_entered_session.initialize.assert_awaited_once()
 
